;
; Copyright 1999 by Konrad Rzeszutek
; Author:    Konrad Rzeszutek 
; File:      deduce-ZOL.lsp
; Date:      Apr 26th 1999
;
; PROLOG:
;
; The world of knowledge representation is a hard and on an unjustiful path.
; Hence, this program tries to easy this task. It does that  by allowing one 
; to check one's ZOL (Zero Order Logic) clauses against ones conclusion. 
; So you can think as this an automated check-machine which would try to
; prove your conclusion. 
;
; This task is accomplished by a method named derivation by resolution 
; refutation. The word derivation in this context really means proof.
; "The proof technique is proof-by-contradiction, whence the term refuation."
; (Bill 2)
;
; To understand the rest of this explanation, one has to understand first the 
; definitions:
;
; * literal is a single propositional symbol P or its negation ~P
; * clause is a disjunction of literals: L1 v L2 v L3 v .. v Ln
; * wff is in conjuctive normal form (CNF) if it is a conjunction of clauses
;   C1 & C2 & .. & Ck
;
; The implementation of wffs is a list of clauses.
; A clause consist of two lists -- one for the negative propositional symbols and another
; list of postive propositional symbols (of course they have to be in CNF -- refer to [1]
; for method of converting)
; Example:  ((A B) (A C))  represents the clause ~A v ~B v A v C
;
;
; The method which uses the derivation by resolution refutation (say that fast)
; automaticly deduces whether some wff W follows from some other set F of wffs
; (Bill 2) and stores that into a set S. This is accomplished by resolution of two clauses, 
; one from the set W and the other from set F. As this tasks is 
; being accomplished for _all_ clauses in F and in W, a proof by
; contradiction is either found or not.
;
; The set of W  do change as the list of currently used premises gets exhausted (W)
;  -- they will be replaced with the resolvent of the n-th clauses with 
; _all_ of the clauses from set S. This motion  tries to form resolvents, resolvent
; of resolvent, etc until a contradiction is found, or if we couldn't find anything
; to resolve from.
;
;
; The method where resolution takes between among two clauses is quite easy. First,
; both clauses have to be in a CNF format (they have to be disjunctions). The method
; finds and removes the negative and positive propositional symbols (P and ~P) from 
; both clauses -- the tautologicals, and the duplicates (P and P).
; This new clause without the tautologicals is now returned.
; Example:
;  C1 = ~A v ~B v A v C      ; ((A B) (A C))
;  C2 = ~C v ~D v A          ; ((C D) (A))
; Resolvent is ~B v ~D       ; ((B D) ())
;  
;
; Note:
;
; To find out if your premises work against your conclusion, first convert those 
; terms into a CNF and then apply this program.
; Example:
; Premises:
;  P => (Q => R)
;  P => Q
; Conclusion:
;  P => (Q & R)
;
; When written as CNF (look on how to do it [1]), the premises become's:
;  ~P v ~ Q v R
;  ~P v Q
; The conclusion becomes:
;  P & (~ Q v ~R)
; which yields two clauses:
;  P
;  ~Q v ~R
;
; So, we would have four diffrent clauses which the program would try to prove:
; ( ((P Q) (R)) ((P) (Q)) (() (P)) ((Q R) ()) )
;
; (simple-level-exhaustion-of '( ((P Q) (R)) ((P) (Q)) (() (P)) ((Q R) ()) ))
; .. would do the refutaion for you,
;
;  
; Works Cited
; [1] Dr. Greene Bill. Homework 4 explanation. 1999
;



; *===========================*
; * Main routine              *
; *===========================*

;
; Simple level exhaustion-of premises.
;
; This neatly algorhytm does automated deduction. This scheme
; produces resolvents of the premises, and resolvents of resolvents,
; etc., which of course create a couple of unneceasary clauses
; before [] pops up. This algorhytm is defined as  simple level
; exhaustion.
;
; We start first with two lists:
;  L0 = premises
;  B0 = premises
; and then for n >= 1
;  Ln = Resoultion (Ln-1, Bn-1)
;  Bn = (Bn-1 U Ln)
;
; This means that for each Bn-1 we would create a list of resolvent
; stored into Ln. So the resolvents would be generated in
; levels L1, L2, L3 ...

(defun simple-level-exhaustion-of (premises)  
  (let(
       (new-clauses  (copy-list  premises))  ;  corresponds to Bn-1
       (body-clauses (copy-list  premises))  ; .. to Ln-1
       (resolvents-that-are-new-clauses (list)); .. to Ln
       (success nil)
       (level 1) 
       (position 1)
       (starting-position-of-new-clauses 0)  ; the position of Bn-1 in Ln-1
       )    
    (dolist (premise premises)
      (setq premise (resolve premise '(()()))) ; sort them
      (print-of premise -1 -1 position)        ; and print them
      (incf position 1))
    (loop
      (format t "LEVEL-~A-RESOLVENTS:~%" level )
      (incf level 1)
      (do*
	  ((index-of-new-clauses 0 (+ 1 index-of-new-clauses)) ; go thru the Bn-1 (onlu its last level) set of 
	   )                                                   ; of clauses and resolve
	  ((or (= (length new-clauses) index-of-new-clauses) success)
	   nil)
	(do*
	    ((index-of-body-clauses 0 (+ 1 index-of-body-clauses)) ; against L1 to Ln-1 (all of its levels).
	     (resolvent (resolve
			 (nth index-of-new-clauses new-clauses)    ; find the new resolvent 
			 (nth index-of-body-clauses body-clauses)) ; of the refuation of two
			(resolve                                   ; clauses
			 (nth index-of-new-clauses new-clauses)
			 (nth index-of-body-clauses body-clauses))))
	    ( (or (= (length body-clauses) index-of-body-clauses) success)
	      nil)
	  (cond	  ( (and ; check to see if [] found
		     (null (first resolvent))
		     (null (second resolvent)))
		    (format t "~4D. [] ~50T[~4D ~4D]~%" position (+ index-of-body-clauses 1) (+ starting-position-of-new-clauses 1))		   
		    (setq success t)  (return))
		  ((or  (duplicates-in resolvent body-clauses) ; check to see if its a duplicate
			(duplicates-in resolvent resolvents-that-are-new-clauses))) ; if so, discard it
		  (t ; otherwise add to the list (to Ln) 
		   (print-of resolvent (+ index-of-body-clauses 1) (+ starting-position-of-new-clauses 1) position)
		   (incf position 1)
		   (setq resolvents-that-are-new-clauses (append resolvents-that-are-new-clauses (list resolvent)))
		   ))))
      (if success (return)) ; if there was a succes or failure -- quit
      (when (and ; if Ln has nothing, then we didn't resolve anything new.
	     (null (first (first resolvents-that-are-new-clauses)))
	     (null (second (first  resolvents-that-are-new-clauses))))
	(format t "~4D. No proof possible." position) ; hence, no proof.
	(return))
      (setq starting-position-of-new-clauses (length body-clauses)) ; change it so we know where Bn-1 is
      (setq body-clauses (nconc body-clauses resolvents-that-are-new-clauses)) ; add Ln to Ln-1
      (setq new-clauses resolvents-that-are-new-clauses) ; copy Ln to Ln-1
      (setq resolvents-that-are-new-clauses nil)
      )))


; *===========================*
; * Useful routines           *
; *===========================*

;
;
; Print a clause with the location of the proof.
;

(defun print-of (clauses parent-one parent-two position)
  
  (format t "~4D. " position)

  (unless (null (first (first clauses))) ; Print the negative first, check to see
    (format t " ~~~A" (first (first clauses))) ; if there is a need to add a 'v'
    (dolist (items (rest (first clauses)))
      (format t " v ~~~A" items)))
  
  (unless (null (first (second clauses))); Print the positive terms.
    (if (not (null (first (first clauses)))) (format t " v"))
    (format t " ~A" (first (second clauses)))
    (dolist (items (rest (second clauses)))
      (format t " v ~A" items)))
  
  (if (= parent-one -1)
      (format t "~50T[premises]~%")
    (format t "~50T[~4D ~4D]~%" parent-one parent-two))
  )

		
;
; 
; Resolves two clauses
;
(defun resolve (clause-A clause-B)

  (let
      ((new-resolvent
	(list                               ; create a new ZOL clause which is a list of two lists.
	                                    ; first list:
	 (sort                              ; concanate the list of negative propo. terms from the clause-A
	  (delete-duplicates                ;  against clause's-B negative propo. terms. After that, delete 
	   (append                          ; delete duplicates and sort the list for easy of comparison
	    (copy-list (first clause-A))   
	    (copy-list (first clause-B))))  
	   #'string-lessp)
	 (sort                              ; second list:
	  (delete-duplicates                ; 
	   (append                          ; the same explanation, but this time the positive propositional
	    (copy-list (second clause-A))   ; terms 
	    (copy-list (second clause-B))))
	  #'string-lessp))))
    (dolist
	(items (intersection                ; for the intersection of positive and negative terms
		(first new-resolvent) (second new-resolvent))) ;  -- find tautologies
      (setf (second new-resolvent) (delete items (second new-resolvent))) ; remove if it exist in negative propositional terms
      (setf (first new-resolvent) (delete items (first new-resolvent)))  ; and as well with the positive proposotional terms
      )
    new-resolvent
))

;
; Looks for duplicates of the 'clause' inside the list of clauses identified as 'of-body'
;
(defun duplicates-in (clause of-body)

  (let 
      ((answer nil))
    (dolist (elements of-body)
      (when (and
	     (equalp (first clause) (first elements))   ; we have to the lists sorted, otherwise
	     (equalp (second clause) (second elements))) ; the lists wouldn't equal.
	(setq answer t)
	(return)))      
    answer
))
